from torch.utils.data import Dataset
import os
from torchvision import transforms
from abc import ABC, abstractmethod
from PIL import Image, ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True
def pil_loader(filename):
    ext = os.path.splitext(filename)[-1]
    ext = ext.lower()
    if ext == '.png' or ext == '.jpeg' or ext == '.ppm' or ext == '.jpg':
        img = Image.open(filename)
        img = img.convert('RGB')
        # img = np.array(img).astype(dtype=np.uint8)
        # img = img[:,:,::-1]  #convert to BGR
        return img
    else:
        raise NotImplementedError('Unsupported file type %s'%ext)

class BaseDataset(Dataset,ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __len__(self):
        return 0

    @abstractmethod
    def __str__(self):
        pass
    
    @staticmethod
    def modify_commandline_options(parser,istrain=False):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        return parser

    def transform_train(self):
        return transforms.Compose([
                transforms.RandomResizedCrop(self.input_size, scale=(0.16, 1), ratio=(0.75, 1.33)),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(mean=self.normal_mean, std=self.normal_std)])

    def transform_validation(self):
        return transforms.Compose([
                transforms.Resize(self.input_size),
                transforms.CenterCrop(self.input_size),
                transforms.ToTensor(),
                transforms.Normalize(mean=self.normal_mean, std=self.normal_std)])



import torch
from torch.optim.optimizer import Optimizer, required

class ours_opt(Optimizer):
	def __init__(self, params, a, b, alpha, lr=required, momentum=0,
				 dampening=0, weight_decay=0, nesterov=False):
		if lr is not required and lr < 0.0:
			raise ValueError("Invalid learning rate: {}".format(lr))
		if momentum < 0.0:
			raise ValueError("Invalid momentum value: {}".format(momentum))
		if weight_decay < 0.0:
			raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
		
		defaults = dict(lr=lr,
						momentum=momentum,
						dampening=dampening,
						weight_decay=weight_decay,
						nesterov=nesterov,
						a=a,
						b=b,
						alpha=alpha)
		if nesterov and (momentum <= 0 or dampening != 0):
			raise ValueError("Nesterov momentum requires a momentum and zero dampening")
		
		super(ours_opt, self).__init__(params, defaults)
	
	def __setstate__(self, state):
		super(ours_opt, self).__setstate__(state)
		for group in self.param_groups:
			group.setdefault('nesterov', False)
			
	def step(self, closure=None):
		"""Performs a single optimization step.
		
		"""
		loss = None
		if closure is not None:
			loss = closure()
		clip_value = 2.0
		for group in self.param_groups:
			weight_decay = group['weight_decay']
			momentum = group['momentum']
			dampening = group['dampening']
			nesterov = group['nesterov']
			
			a = group['a']
			b = group['b']
			alpha = group['alpha']
			
			grad_alpha = torch.clamp(alpha.grad.data, -clip_value, clip_value)
			grad_a = torch.clamp(a.grad.data, -clip_value, clip_value)
			grad_b = torch.clamp(b.grad.data, -clip_value, clip_value)
			
			if weight_decay != 0:
				grad_alpha.add_(weight_decay, alpha.data)
				grad_a.add_(weight_decay, a.data)
				grad_a.add_(weight_decay, b.data)
			
			if momentum != 0:
				param_state = self.state[alpha]
				if 'momentum_buffer' not in param_state:
					buf = param_state['momentum_buffer'] = torch.clone(grad_alpha).detach()
				else:
					buf = param_state['momentum_buffer']
					buf.mul_(momentum).add_(1 - dampening, grad_alpha)
					
				if nesterov:
					grad_alpha = grad_alpha.add(momentum, buf)
				else:
					grad_alpha = buf
				
				param_state = self.state[a]
				if 'momentum_buffer' not in param_state:
					buf = param_state['momentum_buffer'] = torch.clone(grad_a).detach()
				else:
					buf = param_state['momentum_buffer']
					buf.mul_(momentum).add_(1 - dampening, grad_a)
				
				if nesterov:
					grad_a = grad_a.add(momentum, buf)
				else:
					grad_a = buf
				
				param_state = self.state[b]
				if 'momentum_buffer' not in param_state:
					buf = param_state['momentum_buffer'] = torch.clone(grad_b).detach()
				else:
					buf = param_state['momentum_buffer']
					buf.mul_(momentum).add_(1 - dampening, grad_b)
				
				if nesterov:
					grad_b = grad_b.add(momentum, buf)
				else:
					grad_b = buf
				
			# max
			alpha.data.add_(group['lr'], grad_alpha)
			# alpha.data = torch.clamp(alpha.data, -clip_value, clip_value)
			
			# min
			a.data.add_(-group['lr'], grad_a)
			b.data.add_(-group['lr'], grad_b)

			for p in group['params']:
				if p.grad is None:
					continue
				d_p = torch.clamp(p.grad.data , -clip_value, clip_value)
				# d_p = p.grad.data
				if weight_decay != 0:
					d_p.add_(weight_decay, p.data)
				if momentum != 0:
					param_state = self.state[p]
					if 'momentum_buffer' not in param_state:
						buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
					else:
						buf = param_state['momentum_buffer']
						buf.mul_(momentum).add_(1 - dampening, d_p)
					if nesterov:
						d_p = d_p.add(momentum, buf)
					else:
						d_p = buf
				
				p.data.add_(-group['lr'], d_p)
				
		return loss


import torch.nn.functional as F

from .base_model import BaseModel
from .resnet import resnet50


class RetrievalModel(BaseModel):
    def __init__(self, output_channels, freeze=False, freezeBN=False):
        super(RetrievalModel, self).__init__()
        self.output_channels = output_channels
        self.backbone = resnet50(output_channels)

        if freeze:
            self.freeze()

        if freezeBN:
            self.freezeBN()

    def forward(self, x):
        """ Forward

        """
        z = self.backbone(x)
        z = z.view(z.shape[0], -1)
        z = F.normalize(z, dim=1)

        return z

    def param_groups(self, lr=None, lr_fc_mul=1):
        params = list(filter(lambda x: 'fc' not in x[0] and x[1].requires_grad, self.named_parameters()))
        params = [x[1] for x in params]
        fc_params = self.backbone.fc.parameters()

        if len(params):
            if lr is not None:
                return [{'params': params, 'lr': lr}, 
                         {'params': fc_params, 'lr': lr*lr_fc_mul}]
            else:
                return [{'params': params}, {'params': fc_params}]
        else:
            return []



import torch
import torch.nn.functional as F
import numpy as np
import gc


class Evaluator(object):
    def __init__(self):
        self.feats = []
        self.targets = []

    def run_ap(self):
        feats = torch.cat(self.feats, 0).cuda()
        targets = torch.cat(self.targets, 0).cuda()

        i = 0
        gap = 1000
        AP = []

        while i < len(targets):
            sim = torch.mm(feats[i:i+gap], feats.t())
            idx = torch.argsort(-sim)
            tg = targets[i:i+gap].clone()
            tg_gall = targets.clone().view(1,-1).repeat(gap, 1).gather(1, idx)[:, 1:]
            pos_mask = tg_gall == tg.view(-1, 1)

            ## compute AP for each query
            for j in range(len(tg_gall)):
                pos_idx = pos_mask[j].nonzero(as_tuple=False).view(-1) + 1
                ap = (torch.arange(1, len(pos_idx) + 1).cuda() / pos_idx).mean()
                AP.append(ap)

            i += gap

        AP = torch.stack(AP, 0)
        return AP.mean().item()

    def run_rec_at_k(self, k_list=[1, 4, 16, 32]):
        feats = torch.cat(self.feats, 0).cuda()
        targets = torch.cat(self.targets, 0).cuda()

        i = 0
        gap = 1000
        rec_at_k = dict()
        for k in k_list:
            rec_at_k[k] = []

        while i < len(targets):
            sim = torch.mm(feats[i:i+gap], feats.t())
            idx = torch.argsort(-sim)
            tg = targets[i:i+gap].clone()
            tg_gall = targets.clone().view(1,-1).repeat(gap, 1).gather(1, idx)[:, 1:]
            pos_mask = tg_gall == tg.view(-1, 1)

            ## compute recall@k
            for k in k_list:
                rec_at_k[k].append((pos_mask[:, :k].sum(1) > 0).float())

            i += gap

        res = []
        for k in k_list:
            rec_at_k[k] = torch.cat(rec_at_k[k], 0)
            res.append(rec_at_k[k].mean().item())

        if len(k_list) == 1:
            return res[0]

        return res

    def add_batch(self, feat, target):
        self.feats.append(feat.data.cpu())
        self.targets.append(target.data.cpu())

    def reset(self):
        del self.feats
        del self.targets
        gc.collect()
        self.feats = []
        self.targets = []

def _data_preprocess(feat, target):
    if isinstance(feat, torch.Tensor):
        pass
    elif isinstance(feat, np.ndarray):
        feat = torch.from_numpy(feat)
    else:
        raise NotImplementedError

    if isinstance(target, torch.Tensor):
        pass
    elif isinstance(target, np.ndarray):
        target = torch.from_numpy(target)
    else:
        raise NotImplementedError

    feat = F.normalize(feat, p=2, dim=1)
    return feat, target

def AUPRC(feats, targets):
    feats, targets = _data_preprocess(feats, targets)
    evaluator = Evaluator()
    evaluator.add_batch(feats, targets)
    return evaluator.run_ap()

def RecallAtK(feats, targets, k):
    feats, targets = _data_preprocess(feats, targets)
    if isinstance(k, int):
        k = [k]
    evaluator = Evaluator()
    evaluator.add_batch(feats, targets)
    return evaluator.run_rec_at_k(k)


__all__ = [AUPRC, RecallAtK]



import pytest
import sys
sys.path.append("./")
import torch

from XCurve.AUPRC import AUPRC, RecallAtK


@pytest.mark.skip()
def gen_data(emb_dim, num_classes):
    feats = torch.randn((2**14, emb_dim)).numpy()
    targets = torch.randint(0, num_classes, (2**14, 1)).numpy()
    return feats, targets

def test_metrics():

    eps = 1e-4
    for num_classes in [10, 100, 1000]:
        feats, targets = gen_data(128, num_classes)
        auprc = AUPRC(feats, targets)
        rec_at_1_ = RecallAtK(feats, targets, 1)
        rec_at_1, rec_at_10, rec_at_100 = RecallAtK(feats, targets, [1, 10, 100])
        assert 0 < auprc < 1
        assert abs(rec_at_1_ - rec_at_1) < eps


if __name__ == '__main__':
    test_metrics()



__copyright__ = 'XCurve'
__email__ = 'baoshilong@iie.ac.cn'


from .StandardAUROC import SquareAUCLoss, ExpAUCLoss, HingeAUCLoss
from .PartialAUROC import RelaxedPAUCLoss, InsRelaxedPAUCLoss, UnbiasedPAUCLoss
from .AdversarialAUROC import AdvAUROCLoss, RegAdvAUROCLoss, PGDAdversary

def get_losses(args):
    return eval(args.loss_type)(**(args.loss_params))

__all__ = ['SquareAUCLoss', 'ExpAUCLoss', 'HingeAUCLoss',
            'RelaxedPAUCLoss', 'InsRelaxedPAUCLoss', 'UnbiasedPAUCLoss',
            'AdvAUROCLoss', 'RegAdvAUROCLoss', 'PGDAdversary']



import torch
from torch.optim import SGD


class ASGDA(SGD):
    def __init__(self, params=None, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False, init_lr=0.01, hparams=None):
        super(ASGDA, self).__init__(params, momentum, dampening,
                                     weight_decay, nesterov)
        # self.device = torch.device('cuda:0')
        self.device = hparams['device']
        self.nu = torch.tensor(hparams['nu']).to(self.device)
        self.lam = torch.tensor(hparams['lam']).to(self.device)
        self.k = torch.tensor(hparams['k']).to(self.device)
        self.m = torch.tensor(hparams['m']).to(self.device)
        self.c1 = torch.tensor(hparams['c1']).to(self.device)
        self.c2 = torch.tensor(hparams['c2']).to(self.device)
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                state['old_param'] = torch.zeros(p.shape).to(self.device)
                state['old_grad'] = torch.zeros(p.shape).to(self.device)

    @torch.no_grad()
    def step(self, closure=None, pre=False, t=0):
        """Performs a single optimization step.

        Args:
          closure (callable, optional): A closure that reevaluates the model
              and returns the loss.
        """
        # t = torch.tensor(t).to(self.device)
        eta = self.k/torch.pow(self.m+t, 0.333)
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            d_p_list = []
            momentum_buffer_list = []
            old_grad_list = []
            old_param_list = []
            for p in group['params']:
                if p.grad is not None:
                    state = self.state[p]
                    params_with_grad.append(p)
                    if pre:
                        state['old_grad'].zero_()
                        state['old_grad'].add_(p.grad, alpha=1)
                        if 'clip' in group.keys():
                            # if group['name'] == 'g':
                            #     group['clip'][0] = max(
                            #         self.param_groups[1]['params'][1] - 1, -self.param_groups[1]['params'][0])
                            #     d_p_list.append(torch.clip(
                            #         p + self.lam * state['old_param'], group['clip'][0], group['clip'][1]))
                            # else:
                            d_p_list.append(torch.clip(
                            p - self.nu * state['old_param'], group['clip'][0], group['clip'][1]))
                        else:
                            d_p_list.append(p - self.nu * state['old_param'])
                    old_grad_list.append(state['old_grad'])
                    old_param_list.append(state['old_param'])
                    if 'momentum_buffer' not in state:
                        momentum_buffer_list.append(None)
                    else:
                        momentum_buffer_list.append(state['momentum_buffer'])
            if pre:
                sgd(params_with_grad,
                    d_p_list,
                    momentum_buffer_list,
                    eta,
                    weight_decay=group['weight_decay'],
                    momentum=group['momentum'],
                    dampening=group['dampening'],
                    nesterov=group['nesterov'])
                # update momentum_buffers in state
                for p, momentum_buffer, o_p, o_g in zip(params_with_grad, momentum_buffer_list, old_param_list, old_grad_list):
                    state = self.state[p]
                    state['momentum_buffer'] = momentum_buffer
                    state['old_param'] = o_p
                    state['old_grad'] = o_g
            else:
                rho = self.c1 * torch.square(eta)
                xi = self.c2 * torch.square(eta)
                for p, momentum_buffer, o_p, o_g in zip(params_with_grad, momentum_buffer_list, old_param_list, old_grad_list):
                    state = self.state[p]
                    state['momentum_buffer'] = momentum_buffer
                    state['old_grad'] = o_g
                    if group['name'] == 'g':
                        all_grad = p.grad + (1 - xi) * \
                            (state['old_param'] - o_g)
                    else:
                        all_grad = p.grad + (1 - rho) * \
                            (state['old_param'] - o_g)
                    state['old_param'].zero_()
                    state['old_param'].add_(all_grad, alpha=1)

        return loss


def sgd(params,
        d_p_list,
        momentum_buffer_list,
        lr,
        weight_decay,
        momentum,
        dampening,
        nesterov):
    r"""Functional API that performs SGD algorithm computation.

    See :class:`~torch.optim.SGD` for details.
    """

    for i, param in enumerate(params):

        d_p = d_p_list[i]
        if weight_decay != 0:
            d_p = d_p.add(param, alpha=weight_decay)

        if momentum != 0:
            buf = momentum_buffer_list[i]

            if buf is None:
                buf = torch.clone(d_p).detach()
                momentum_buffer_list[i] = buf
            else:
                buf.mul_(momentum).add_(d_p, alpha=1 - dampening)

            if nesterov:
                d_p = d_p.add(buf, alpha=momentum)
            else:
                d_p = buf
        param.mul_(1-lr)
        param.add_(lr*d_p, alpha=1)



from .loss_warpper import LossWarpper

""" SOPRC

Paper:
`Exploring the Algorithm-Dependent Generalization of AUPRC Optimization with List Stability` - https://proceedings.neurips.cc/paper_files/paper/2022/file/b5dc49f44db2fadc5c4d717c57f4a424-Paper-Conference.pdf
@article{wen2022exploring,
  title={Exploring the Algorithm-Dependent Generalization of AUPRC Optimization with List Stability},
  author={Wen, Peisong and Xu, Qianqian and Yang, Zhiyong and He, Yuan and Huang, Qingming},
  journal={Advances in Neural Information Processing Systems},
  volume={35},
  pages={28335--28349},
  year={2022}
}

Original code and weights from: https://github.com/KID-7391/SOPRC
Copyright (c) Inst. of Computing Tech., CAS.
All rights reserved.
This source code is licensed under the MIT license

"""

def ListStableAUPRC(
        tau1=0.1, 
        tau2=0.001,
        beta=0.001,
        prior_mul=0.1,
        num_sample_per_id=4,
        var_reg_weight_pos=5,
        var_reg_weight_neg=1
    ):

    """
    Args:
        tau1 (float): Control the surrogate loss of pos-neg pairs. See \tau_1 in Eq.(7).
        tau2 (float): Control the surrogate loss of pos-pos pairs. See \tau_2 in Eq.(7).
        beta (float): Control the exponential moving average. See \beta in Eq.(10).
        prior_mul (float): Imbalance ratio of the id with most positive examples.
        num_sample_per_id (int): Number of examples for each id.
        var_reg_weight_pos (float): Weight of the variance regular term w.r.t. positive examples.
        var_reg_weight_neg (float): Weight of the variance regular term w.r.t. negative examples.
    """

    args = {
        'soprc': {
            'loss_weight': 1,
            'loss_param': {
                'tau': [tau1, tau2],
                'beta': beta,
                'prior_mul': prior_mul,
                'num_sample_per_id': num_sample_per_id
            }
        },
        'variance': {
            'loss_weight': 1,
            'loss_param': {
                'thres': [var_reg_weight_pos, var_reg_weight_neg],
                'num_sample_per_id': num_sample_per_id
            }
        }
    }
    return LossWarpper(args)

__all__ = [ListStableAUPRC]



import torch

lower_limit, upper_limit = 0.0, 1.0

def normalize(X):
    # global mu, std
    # return (X - mu)/std
    return X

def clamp(X, lower_limit, upper_limit):
    return torch.max(torch.min(X, upper_limit), lower_limit)

def PGDAdversary(model, X, y, criterion, epsilon=8.0/255, 
		        alpha=2.0/255, 
		        attack_iters=10, 
		        restarts=1, 
		        norm='linf'): 
	# model.eval()
	max_loss = torch.zeros(y.shape[0]).cuda() 
	max_delta = torch.zeros_like(X).cuda() 
	for _ in range(restarts): 
		delta = torch.zeros_like(X) 
		delta.uniform_(-epsilon, epsilon)
		delta = clamp(delta, lower_limit - X, upper_limit - X)
		delta.requires_grad = True
        
		for _ in range(attack_iters):
			output = model(normalize(X + delta)).view_as(y)
				
			index = slice(None, None, None)
				
			loss = criterion(output, y)
			loss.backward()
			grad = delta.grad.detach()
				
			d = delta[index, :, :, :]
			g = grad[index, :, :, :]
			x = X[index, :, :, :]

			if norm == 'linf':
				d = torch.clamp(d + alpha * torch.sign(g), 
									min=-epsilon, max=epsilon)
			else: 
				g_norm = torch.norm(g.view(g.shape[0],-1),dim=1).view(-1,1,1,1)
				scaled_g = g/(g_norm + 1e-10)
				d = (d + scaled_g * alpha).view(d.size(0),
													-1).renorm(p=2,
							     								dim=0,
																maxnorm=epsilon).view_as(d)
			d = clamp(d, lower_limit - x, upper_limit - x)
			delta.data[index, :, :, :] = d
			delta.grad.zero_()
			all_loss = criterion(model(normalize(X + delta)).view_as(y), y)
				
		max_delta[all_loss >= max_loss] = delta.detach()[all_loss >= max_loss]
		max_loss = torch.max(max_loss, all_loss)
	
	# model.train()
	return max_delta


import torch

def evaluate(output, target, k_list=(1, )):
    return TopkAcc(output, target, k_list), AUTKC(output, target, k_list)


def TopkAcc(y_pred, y_true, k_list=(1, )):
    """
    Computes the top-k accuracy for each k in the specified k-list.
    :param y_pred: the predictions of a multiclass model (batch_size, n_classes)
    :param y_true: the ground-truth labels (batch_size, )
    :param k_list: the list of the specified k
    :return: top-k accuracy
    :reference: ...
    """
    with torch.no_grad():
        maxk = max(k_list)
        batch_size = y_true.size(0)

        _, pred = y_pred.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(y_true.view(1, -1).expand_as(pred))

        res = []
        for k in k_list:
            correct_k = correct[:k].contiguous().view(-1).float().sum(0)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res

def AUTKC(y_pred, y_true, k_list=(1, )):
    with torch.no_grad():
        s_y = y_pred.gather(-1, y_true.view(-1, 1))
        s_top, _ = y_pred.topk(max(k_list) + 1, 1, True, True)
        tmp = s_y > s_top

        res = []
        for k in k_list:
            tmp_k = tmp[:, :k + 1]
            autkc_k = torch.sum(tmp_k.float(), dim=-1) / k
            res.append(autkc_k)
        return [_.mean() * 100 for _ in res]


import os
import os.path as osp
import torch
import random
import copy

from .base_dataset import BaseDataset
from .base_dataset import pil_loader


class RetrievalDataset(BaseDataset):
    def __init__(self, data_dir, list_dir, subset, input_size, batchsize, num_sample_per_id, normal_mean=[0.485, 0.456, 0.406], normal_std=[0.229, 0.224, 0.225], split='train', **kwargs):
        super().__init__()

        self.input_size = input_size
        self.batchsize = batchsize
        self.num_sample_per_id = num_sample_per_id
        self.normal_mean = normal_mean
        self.normal_std = normal_std

        if subset is None:
            subset = kwargs['dataset_' + split]

        self.data_dir = data_dir
        self.split = split
        if split == 'train':
            _data_list = os.path.join(list_dir, subset + '.txt')
            self.transform = self.transform_train()
        elif split == 'val':
            _data_list = os.path.join(list_dir, subset + '.txt')
            self.transform = self.transform_validation()
        elif split == 'test':
            _data_list = os.path.join(list_dir, subset + '.txt')
            self.transform = self.transform_validation()
        else:
            raise ValueError
        self._data_list = _data_list
        self.split = split

        with open(_data_list) as f:
            lines = f.read().splitlines()[1:]

        self.metas = []
        id_mp = {}
        for i, line in enumerate(lines):
            id_ = int(line.split(' ')[1])

            if not id_ in id_mp.keys():
                id_mp[id_] = len(id_mp)
                self.metas.append([])

            id_ = id_mp[id_]
            path = osp.join(self.data_dir, line.split(' ')[-1])
            self.metas[id_].append((path, id_))

        cnt_per_id = [len(i) for i in self.metas]
        max_cnt_per_id = max(cnt_per_id)
        self.cnt_per_id = cnt_per_id

        for id_ in range(len(self.metas)):
            for i in range(len(self.metas[id_])):
                self.metas[id_][i] = tuple(list(self.metas[id_][i]) + \
                    [cnt_per_id[id_] / max_cnt_per_id])

        self.reset()
        print('%s set has %d samples per epoch' % (self.split, self.__len__()))

    def reset(self):
        print('\nshuffling data...')
        datalist = []

        metas = copy.deepcopy(self.metas)
        num_classes = len(metas)
        for i in range(num_classes):
            random.shuffle(metas[i])

        bs = self.batchsize
        ns = self.num_sample_per_id

        classes = [i for i in range(num_classes)]
        random.shuffle(classes)

        mini_batch = []
        while True:
            for clss in classes:
                if len(metas[clss]) >= ns and len(mini_batch) < bs:
                    mini_batch += metas[clss][:ns]
                    metas[clss] = metas[clss][ns:]

                if len(mini_batch) == bs:
                    break

            if len(mini_batch) == bs:
                datalist.append(mini_batch)
                mini_batch = []
            else:
                break

        random.shuffle(datalist)
        self.datalist = []
        for i in datalist:
            self.datalist += i

    def __len__(self):
        return len(self.datalist)

    def __str__(self):
        return self.data_dir + '  split=' + str(self.split)

    def get_cnt_per_id(self):
        return self.cnt_per_id

    def __getitem__(self, idx):
        img_filename = self.datalist[idx][0]
        img = pil_loader(img_filename)
        img = self.transform(img)
        
        return img, torch.tensor([int(self.datalist[idx][1])])



import pytest
import torch
import random
import sys
from torch.utils.data import DataLoader
sys.path.append("./")

from XCurve.AUPRC import (RetrievalDataset,\
     RetrievalModel, ListStableAUPRC, DefaultLossCfg)


@pytest.mark.skip()
def load_cfg(dataset):
    if dataset == 'inat':
        from XCurve.AUPRC import DefaultInatDatasetCfg as DefaultDatasetCfg
    elif dataset == 'sop':
        from XCurve.AUPRC import DefaultSOPDatasetCfg as DefaultDatasetCfg
    elif dataset == 'vehid':
        from XCurve.AUPRC import DefaultVehIDDatasetCfg as DefaultDatasetCfg
    else:
        assert NotImplementedError

    return DefaultDatasetCfg

def test_losses_decrese():
    '''
        check if the loss can drop normally
    '''
    for dataset_name in ['inat', 'sop', 'vehid']:
        args = load_cfg(dataset_name)
        dataset = RetrievalDataset(**args, split='val')
        dataloader = DataLoader(
            dataset,
            batch_size=args.batchsize,
            num_workers=4
        )

        model = RetrievalModel(512).cuda()
        criterion = ListStableAUPRC(**DefaultLossCfg)
        criterion.update_cnt_per_id(dataset.get_cnt_per_id())

        optimizer = torch.optim.SGD(model.param_groups(), lr=0.001)
        train_loss = []
        for i, (images, targets) in enumerate(dataloader):
            images = images.cuda()
            targets = targets.cuda()
            feats = model(images)
            loss = criterion(feats, targets)
            train_loss.append(loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if (i+1)%100 == 0:
                break

        assert sum(train_loss[:10]) > sum(train_loss[-10:]), train_loss


if __name__ == '__main__':
    test_losses_decrese()



from tokenize import group
import packaging
import torch
from torch.optim.optimizer import Optimizer, required

class AdvAUCOptimizer(Optimizer):
    def __init__(self, params, a, b, alpha, lr=required, momentum=0.9,
                 dampening=0, weight_decay=0, nesterov=False):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        
        defaults = dict(lr=lr,
                        momentum=momentum,
                        dampening=dampening,
                        weight_decay=weight_decay,
                        nesterov=nesterov,
                        a=a,
                        b=b,
                        alpha=alpha)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        
        super(AdvAUCOptimizer, self).__init__(params, defaults)
    
    def __setstate__(self, state):
        super(AdvAUCOptimizer, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)
            
    def step(self, closure=None):
        """Performs a single optimization step.
        
        """
        loss = None
        if closure is not None:
            loss = closure()
        clip_value = 1.0
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            
            a = group['a']
            b = group['b']
            alpha = group['alpha']
            
            grad_alpha = torch.clamp(alpha.grad.data, -clip_value, clip_value)
            grad_a = torch.clamp(a.grad.data, -clip_value, clip_value)
            grad_b = torch.clamp(b.grad.data, -clip_value, clip_value)
            
            if weight_decay != 0:
                grad_alpha.add_(weight_decay, alpha.data)
                grad_a.add_(weight_decay, a.data)
                grad_a.add_(weight_decay, b.data)
            
            if momentum != 0:
                param_state = self.state[alpha]
                if 'momentum_buffer' not in param_state:
                    buf = param_state['momentum_buffer'] = torch.clone(grad_alpha).detach()
                else:
                    buf = param_state['momentum_buffer']
                    buf.mul_(momentum).add_(1 - dampening, grad_alpha)
                    
                if nesterov:
                    grad_alpha = grad_alpha.add(momentum, buf)
                else:
                    grad_alpha = buf
                
                param_state = self.state[a]
                if 'momentum_buffer' not in param_state:
                    buf = param_state['momentum_buffer'] = torch.clone(grad_a).detach()
                else:
                    buf = param_state['momentum_buffer']
                    buf.mul_(momentum).add_(1 - dampening, grad_a)
                
                if nesterov:
                    grad_a = grad_a.add(momentum, buf)
                else:
                    grad_a = buf
                
                param_state = self.state[b]
                if 'momentum_buffer' not in param_state:
                    buf = param_state['momentum_buffer'] = torch.clone(grad_b).detach()
                else:
                    buf = param_state['momentum_buffer']
                    buf.mul_(momentum).add_(1 - dampening, grad_b)
                
                if nesterov:
                    grad_b = grad_b.add(momentum, buf)
                else:
                    grad_b = buf
                
            # max
            alpha.data.add_(group['lr'], grad_alpha)
            # min
            a.data.add_(-group['lr'], grad_a)
            b.data.add_(-group['lr'], grad_b)

            # clip
            # a.data = torch.clamp(a.data, 0.0, 1.0)
            # b.data = torch.clamp(b.data, 0.0, 1.0)
            # alpha.data = torch.clamp(alpha.data, max(b.data-1, -a.data), 1)
            
            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = torch.clamp(p.grad.data , -clip_value, clip_value)
                # d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf
                
                p.data.add_(-group['lr'], d_p)
                
        return loss

class RegAdvAUCOptimizer(Optimizer):
    def __init__(self, params, a, b, alpha, lambda1, lambda2, c1=1, c2=1, gamma=1, lamda=1 ,lr=required):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        
        defaults = dict(lr=lr, c1=c1, c2=c2, gamma=gamma, lamda=lamda,
                        a=a,
                        b=b,
                        alpha=alpha,
                        lambda1=lambda1,
                        lambda2=lambda2)
        super(RegAdvAUCOptimizer, self).__init__(params, defaults)
    
    def __setstate__(self, state):
        super(RegAdvAUCOptimizer, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)
    
    def step(self, closure=None):
        """Performs a single optimization step.
        """
        loss = None
        if closure is not None:
            loss = closure()
        clip_value = 1.0
        for group in self.param_groups:
            a = group['a']
            b = group['b']
            alpha = group['alpha']
            lambda1 = group['lambda1']
            lambda2 = group['lambda2']
            # record the old alpha, lambda1, lambda2, a, b, w
            self.a_old, self.b_old, self.alpha_old = torch.clone(a.data).detach(), torch.clone(b.data).detach(), torch.clone(alpha.data).detach()
            self.lambda1_old, self.lambda2_old = torch.clone(lambda1.data).detach(), torch.clone(lambda2.data).detach()
            # get the learning rate
            # c1 = group['c1']
            # c2 = group['c2']
            lr = group['lr']
            gamma = group['gamma']
            lamda = group['lamda']
            
            # alpha
            # grad_alpha = alpha.grad.data
            param_state = self.state[alpha]
            if 'momentum_buffer' not in param_state:
                buf = param_state['momentum_buffer'] = 0
            else:
                buf = param_state['momentum_buffer']
            # what is the gamma and lambda in line 3 and line 5? 
            alpha.data.add_(lamda, buf)
            alpha.data = self.alpha_old + lr * (alpha.data - self.alpha_old)
            alpha.data = torch.clamp(alpha.data, -1.0, 1.0)
            # min
            # a
            # grad_a = a.grad.data
            param_state = self.state[a]
            if 'momentum_buffer' not in param_state:
                buf = param_state['momentum_buffer'] = 0
            else:
                buf = param_state['momentum_buffer']
            a.data.add_(-gamma, buf)
            a.data = self.a_old + lr * (a.data - self.a_old)
            a.data = torch.clamp(a.data, 0, 1.0)
            # b
            # grad_b = b.grad.data
            param_state = self.state[b]
            if 'momentum_buffer' not in param_state:
                buf = param_state['momentum_buffer'] = 0
            else:
                buf = param_state['momentum_buffer']
            b.data.add_(-gamma, buf)
            b.data = self.b_old + lr * (b.data - self.b_old)
            b.data = torch.clamp(b.data, 0, 1.0)
            # lambda1
            # grad_lambda1 = lambda1.grad.data
            param_state = self.state[lambda1]
            if 'momentum_buffer' not in param_state:
                buf = param_state['momentum_buffer'] = 0
            else:
                buf = param_state['momentum_buffer']
            lambda1.data.add_(-gamma, buf)
            # if lambda1.data<0:
            #     lambda1.data = lambda1.data-lambda1.data
            lambda1.data = torch.clamp(lambda1.data, 0.0, 5.0)
            lambda1.data = self.lambda1_old + lr * (lambda1.data - self.lambda1_old)
            
            # lambda2
            # grad_lambda2 = lambda2.grad.data
            param_state = self.state[lambda2]
            if 'momentum_buffer' not in param_state:
                buf = param_state['momentum_buffer'] = 0
            else:
                buf = param_state['momentum_buffer']
            lambda2.data.add_(-gamma, buf)
            # lambda2 = torch.clamp(b.data, 0, 1.0)
            # if lambda2.data<0:
            #     lambda1.data = lambda1.data-lambda1.data
            lambda1.data = torch.clamp(lambda1.data, 0.0, 5.0)
            lambda2.data = self.lambda2_old + lr * (lambda2.data - self.lambda2_old)
            
            # params
            # self.p_old = group['params']
            for p in group['params']:
                if p.grad is None:
                    continue
                # d_p = torch.clamp(p.grad.data , -clip_value, clip_value)
                # d_p = p.grad.data
                p_old = torch.clone(p.data).detach()
                param_state = self.state[p]
                if 'momentum_buffer' not in param_state:
                    buf = param_state['momentum_buffer'] = 0
                else:
                    buf = param_state['momentum_buffer']
                
                p.data.add_(-gamma, buf)
                p.data = p_old + lr * (p.data - p_old)
                
        return loss
    
    # before step, we need to do; which is use grad of z_t from disk
    # opt.zero_grad()
    # loss.backward()
    # opt.record_grad()
    def record_grad(self):
        for group in self.param_groups:
            a = group['a']
            b = group['b']
            alpha = group['alpha']
            lambda1 = group['lambda1']
            lambda2 = group['lambda2']
            
            param_state = self.state[a]
            param_state['grad_buffer'] = torch.clone(a.grad.data).detach()
            param_state = self.state[b]
            param_state['grad_buffer'] = torch.clone(b.grad.data).detach()
            param_state = self.state[alpha]
            param_state['grad_buffer'] = torch.clone(alpha.grad.data).detach()
            param_state = self.state[lambda1]
            param_state['grad_buffer'] = torch.clone(lambda1.grad.data).detach()
            param_state = self.state[lambda2]
            param_state['grad_buffer'] = torch.clone(lambda2.grad.data).detach()
            
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                param_state['grad_buffer'] = torch.clone(p.grad.data).detach()
    
    # after step()
    # opt.zero_grad()
    # loss.backward()
    # opt.updata_momentum()
    def updata_momentum(self):
        for group in self.param_groups:
            a = group['a']
            b = group['b']
            alpha = group['alpha']
            lambda1 = group['lambda1']
            lambda2 = group['lambda2']
            
            c1 = group['c1']
            c2 = group['c2']
            lr = group['lr']
            beta1 = 1 - c1 * lr * lr
            beta2 = 1 - c2 * lr * lr
            
            param_state = self.state[a]
            param_state['momentum_buffer'] = beta1 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(a.grad.data).detach()
            
            param_state = self.state[b]
            param_state['momentum_buffer'] = beta1 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(b.grad.data).detach()
            
            param_state = self.state[lambda1]
            param_state['momentum_buffer'] = beta1 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(lambda1.grad.data).detach()
            
            param_state = self.state[lambda2]
            param_state['momentum_buffer'] = beta1 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(lambda2.grad.data).detach()
            
            param_state = self.state[alpha]
            param_state['momentum_buffer'] = beta2 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(alpha.grad.data).detach()
            
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                param_state['momentum_buffer'] = beta1 * (param_state['momentum_buffer'] - param_state['grad_buffer']) + torch.clone(p.grad.data).detach()



import torch
from abc import abstractmethod

from .base_loss import BaseLoss


class PairLoss(BaseLoss):

    def __init__(self, num_sample_per_id, thres=None, norm=False, **kwargs):
        super(PairLoss, self).__init__()
        self.num_sample_per_id = num_sample_per_id
        self.thres = thres
        self.norm = norm

    def _check_input(self, targets):
        batch_size = targets.shape[0]
        targets = targets.view(
            batch_size // self.num_sample_per_id,
            self.num_sample_per_id
        )
        diff = targets - targets[:, 0].unsqueeze(1)
        assert diff.sum() == 0
        assert (targets == targets[0]).sum() == self.num_sample_per_id

    @abstractmethod
    def _criterion(self, x):
        pass

    def forward(self, feats, targets):
        bs = targets.shape[0]
        nc = bs // self.num_sample_per_id
        ns = self.num_sample_per_id
        d = feats.shape[-1]

        self._check_input(targets)

        mask = torch.block_diag(*([torch.ones(ns, ns) + torch.eye(ns)]*nc)).cuda()

        sim = torch.mm(feats, feats.t())
        sim_pos = torch.index_select(sim.view(-1), 0, mask.view(-1).eq(1).nonzero().view(-1))
        sim_neg = torch.index_select(sim.view(-1), 0, mask.view(-1).eq(0).nonzero().view(-1))

        loss = self._criterion(sim_pos.view(bs, ns - 1), sim_neg.view(bs, (nc - 1) * ns))

        return loss


class VarianceLoss(PairLoss):
    def __init__(self, inter_clss=False, **kwargs):
        super(VarianceLoss, self).__init__(**kwargs)
        self.inter_clss = inter_clss

    def _criterion(self, sim_pos, sim_neg):
        assert len(sim_pos.shape) == 2 and len(sim_neg.shape) == 2
        assert sim_pos.shape[0] == sim_neg.shape[0]

        mean_pos = sim_pos.mean()
        mean_neg = sim_neg.mean()
        target_pos = mean_pos
        target_neg = mean_neg
        diff_pos = (target_pos - sim_pos.view(-1))
        diff_neg = (sim_neg.view(-1) - target_neg)

        diff_pos = torch.index_select(diff_pos, 0, (diff_pos > 0).nonzero().view(-1))
        diff_neg = torch.index_select(diff_neg, 0, (diff_neg > 0).nonzero().view(-1))

        if len(diff_pos) > 0:
            loss_pos = (diff_pos**2).mean()
        else:
            loss_pos = torch.tensor([.0]).cuda()
        if len(diff_neg) > 0:
            loss_neg = (diff_neg**2).mean()
        else:
            loss_neg = torch.tensor([.0]).cuda()
        loss = self.thres[0] * loss_pos + self.thres[1] * loss_neg

        return loss.mean()



__copyright__ = 'Shilong Bao'
__email__ = 'baoshilong@iie.ac.cn'

from .minmax_opt import SGD4MinMaxPAUC
from .ASGDA import ASGDA
from torch.optim import SGD
from torch.optim import Adadelta
from torch.optim import Adagrad
from torch.optim import Adam
from torch.optim import AdamW
from .AdvAUC_opt import AdvAUCOptimizer, RegAdvAUCOptimizer

__all__ = [SGD4MinMaxPAUC, ASGDA, SGD, Adadelta, Adagrad,
            Adam, AdamW, AdvAUCOptimizer, RegAdvAUCOptimizer]


from .pair_loss import VarianceLoss
from .soprc import SOPRC


loss_dict = {
    'variance': VarianceLoss,
    'soprc': SOPRC
}


class LossWarpper(object):
    def __init__(self, args):
        self.args = args
        self.criterion = []
        for loss_name in args.keys():
            self.criterion.append(
                {
                    'loss_fn': loss_dict[loss_name](**args[loss_name]['loss_param']),
                    'loss_name': loss_name,
                    'loss_weight': args[loss_name].get('loss_weight', 1.0)
                }
            )

    def update_cnt_per_id(self, cnt_per_id):
        for i in range(len(self.criterion)):
            if isinstance(self.criterion[i]['loss_fn'], SOPRC):
                self.criterion[i]['loss_fn'].update_cnt_per_id(cnt_per_id)

    def __call__(self, feats, targets):
        loss = 0
        for cri in self.criterion:
            loss += cri['loss_weight'] * cri['loss_fn'](feats, targets)

        return loss

    def param_groups(self, lr=None, **kwargs):
        params = []
        for cri in self.criterion:
            params += cri['loss_fn'].param_groups(lr, **kwargs)
        return params



#coding=utf-8
import torch
import torchvision
import torchvision.transforms as transforms
import numpy as np
import torchvision.datasets as datasets
# from sampler import *

class IMBALANCECIFAR10(torchvision.datasets.CIFAR10):
    cls_num = 10
    def __init__(self, root, imb_type='exp', imb_factor=0.01, rand_number=0, train=True,
                 transform=None, target_transform=None,
                 download=False):
        super(IMBALANCECIFAR10, self).__init__(root, train, transform, target_transform, download)
        np.random.seed(rand_number)
        img_num_list = self.get_img_num_per_cls(self.cls_num, imb_type, imb_factor)
        self.phat = 0.1
        self.gen_imbalanced_data(img_num_list)

        # print(type(self.data.shape))
        
        
    def get_img_num_per_cls(self, cls_num, imb_type, imb_factor):
        img_max = len(self.data) / cls_num
        img_num_per_cls = []
        if imb_type == 'exp':
            for cls_idx in range(cls_num):
                num = img_max * (imb_factor**(cls_idx / (cls_num - 1.0)))
                img_num_per_cls.append(int(num))
        elif imb_type == 'step':
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max))
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max * imb_factor))
        else:
            img_num_per_cls.extend([int(img_max)] * cls_num)
        return img_num_per_cls

    def gen_imbalanced_data(self, img_num_per_cls):
        new_data = []
        new_targets = []
        targets_np = np.array(self.targets, dtype=np.int64)
        classes = np.unique(targets_np)
        # np.random.shuffle(classes)
        self.num_per_cls_dict = dict()
        for the_class, the_img_num in zip(classes, img_num_per_cls):
            self.num_per_cls_dict[the_class] = the_img_num
            idx = np.where(targets_np == the_class)[0]
            np.random.shuffle(idx)
            selec_idx = idx[:the_img_num]
            new_data.append(self.data[selec_idx, ...])
            new_targets.extend([the_class, ] * the_img_num)
        new_data = np.vstack(new_data)
        new_targets = self.get_two_class(new_targets)
        self.data = new_data
        self.targets = new_targets
        
    def get_cls_num_list(self):
        cls_num_list = []
        for i in range(self.cls_num):
            cls_num_list.append(self.num_per_cls_dict[i])
        return cls_num_list

    def get_two_class(self, Y):
        Y = np.array(Y)
        # for i in range(10):
        #     print(i, len(np.where(Y ==i)[0]))
        loc_0 = np.where(Y <= (self.cls_num/2-1))[0]
        loc_1 = np.where(Y > (self.cls_num/2-1))[0]
        Y[loc_1] = 1
        Y[loc_0] = 0
        # for i in range(2):
        #     print(i, len(np.where(Y == i)[0]))
        self.phat = len(np.where(Y == 1)[0])/len(Y)
        return Y.tolist()

class IMBALANCECIFAR100(IMBALANCECIFAR10):
    """`CIFAR100 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    This is a subclass of the `CIFAR10` Dataset.
    """
    base_folder = 'cifar-100-python'
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = 'eb9058c3a382ffc7106e4002c42a8d85'
    train_list = [
        ['train', '16019d7e3df5f24257cddd939b257f8d'],
    ]

    test_list = [
        ['test', 'f0ef6b0ae62326f3e7ffdfab6717acfc'],
    ]
    meta = {
        'filename': 'meta',
        'key': 'fine_label_names',
        'md5': '7973b15100ade9c7d40fb424638fde48',
    }
    cls_num = 100

    def get_two_class(self, Y):
        Y = np.array(Y)
        # for i in range(10):
        #     print(i, len(np.where(Y ==i)[0]))
        loc_0 = np.where(Y <= (self.cls_num/2-1))[0]
        loc_1 = np.where(Y > (self.cls_num/2-1))[0]
        Y[loc_1] = 1
        Y[loc_0] = 0
        print('positive sample', len(loc_1))
        print('negative sample', len(loc_0))
        self.phat = len(np.where(Y == 1)[0]) / len(Y)
        return Y.tolist()


# if __name__ == '__main__':
#     transform_train = transforms.Compose([
#         transforms.RandomCrop(32, padding=4),
#         transforms.RandomHorizontalFlip(),
#         transforms.ToTensor(),
#         # transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
#     ])

#     train_dataset = IMBALANCECIFAR100(root='../', train=True, download=True, transform=transform_train)

#     train_sampler = StratifiedSampler(class_vector=train_dataset.targets,
#                                       batch_size=128,
#                                       rpos=1,
#                                       rneg=9)

#     train_loader = torch.utils.data.DataLoader(
#         train_dataset, batch_size=128, shuffle=(train_sampler is None),
#         num_workers=2,
#         # pin_memory=True,
#         sampler=train_sampler)


#     for i, (x, y) in enumerate(train_loader):
#         print(i, x.min(), x.max(), x.mean(), x.std())



import torch
import torch.nn as nn

from abc import abstractmethod


class BaseLoss(nn.Module):
    def __init__(self):
        super(BaseLoss, self).__init__()
        pass

    def feat2sim(self, feats, targets):
        x = torch.mm(feats, feats.t()) # N*N
        label = (targets.view(-1, 1) == targets.view(1, -1)).long()
        return x, label

    @abstractmethod
    def _preprocess(self):
        pass

    @abstractmethod
    def forward(self):
        pass

    def param_groups(self, lr=None, **kwargs):
        params = list(filter(lambda x:x.requires_grad, self.parameters()))
        if len(params):
            if lr is not None:
                return [{'params': params, 'lr': lr}]
            else:
                return [{'params': params}]
        else:
            return []



__copyright__ = 'XCurve'
__email__ = 'baoshilong@iie.ac.cn'


from .AdversarialAUROC import AdvAUROCLoss, RegAdvAUROCLoss
from .attacker import PGDAdversary
__all__ = ['AdvAUROCLoss', 'RegAdvAUROCLoss', 'PGDAdversary']



import torch
import math

from .base_loss import BaseLoss


class ListStableAUPRCLoss(BaseLoss):

    def __init__(self, num_sample_per_id, tau, beta, prior_mul=0.1, **kwargs):
        super(ListStableAUPRCLoss, self).__init__()

        self.num_sample_per_id = num_sample_per_id
        self.temp = tau
        self.beta = beta
        self.prior_mul = prior_mul
        self.num_v = 100
        self.v = None
        self.alpha = None

    def update_cnt_per_id(self, cnt_per_id):
        if self.prior_mul < 0:
            self.prior = None
            return

        self.cnt_per_id = cnt_per_id
        self.v = torch.stack([torch.arange(self.num_v) / self.num_v for _ in range(len(cnt_per_id))]).cuda()
        self.prior = torch.tensor(cnt_per_id).float().cuda()
        self.prior *= self.prior_mul / self.prior.max()

    def _check_input(self, targets):
        assert self.prior_mul < 0 or self.v is not None
        batch_size = targets.shape[0]
        targets = targets.view(
            batch_size // self.num_sample_per_id,
            self.num_sample_per_id
        )
        diff = targets - targets[:, 0].unsqueeze(1)
        assert diff.sum() == 0
        assert (targets == targets[0]).sum() == self.num_sample_per_id

    def _surrogate_fn_pn(self, x):
        temp = self.temp[0]
        x = torch.clamp(x, max=temp)
        return torch.where(
            x >= 0,
            (x / temp - 1)**2,
            (-2 * x / temp + 1)
        )

    def _surrogate_fn_pp(self, x):
        temp = self.temp[1]
        x = torch.clamp(x, max=0)
        return 2 / (1 + torch.exp(x / temp)) - 1

    def forward(self, feats, targets):
        ## preprocess input 
        targets = targets.view(-1)
        batch_size = targets.shape[0]
        nc = batch_size // self.num_sample_per_id
        ns = self.num_sample_per_id
        self._check_input(targets)

        ## compute similarity scores
        mask = torch.block_diag(*([torch.ones(ns, ns) + torch.eye(ns)]*nc)).cuda()
        sim = torch.mm(feats, feats.t())
        sim_pos = torch.index_select(sim.view(-1), 0, mask.view(-1).eq(1).nonzero().view(-1))
        sim_pos = sim_pos.view(batch_size, ns - 1, 1)
        sim_pos = torch.sort(sim_pos, 1)[0]
        sim_neg = torch.index_select(sim.view(-1), 0, mask.view(-1).eq(0).nonzero().view(-1))
        sim_neg = sim_neg.view(batch_size, 1, (nc - 1)*ns)

        ## l_1
        loss_pn = self._surrogate_fn_pn(sim_pos - sim_neg).mean(-1)

        ## update v by score intp
        with torch.no_grad():
            beta = self.beta
            if self.alpha is None:
                len_intv = math.ceil(self.num_v / (ns-1))
                self.alpha = torch.arange(len_intv).view(1,1,-1) / len_intv
                self.alpha = self.alpha.cuda()

            score = sim_pos.clone()
            score_left = torch.cat([
                torch.clamp(2*score[:, 0] - score[:, 1], min=-1).unsqueeze(1),
                score[:, :-1]
            ], dim=1)
            v_left = (score - score_left) * self.alpha + score_left

            score_right = torch.cat([
                score[:, 1:],
                torch.clamp(2*score[:, -1] - score[:, -2], max=1).unsqueeze(1)
            ], dim=1)
            v_right = (score_right - score) * self.alpha + score

            self.v[targets] = (1-beta) * self.v[targets] + beta / 2 * (v_left + v_right).view(batch_size, -1)[:, :self.num_v]

            ## l_2
            v = torch.index_select(self.v, 0, targets)

            loss_pv = self._surrogate_fn_pp(sim_pos - v.view(batch_size, 1, -1))
            loss_pv = (1 + loss_pv.sum(-1)) / (loss_pv.shape[-1] + 1)

        ## biased estimation of auprc
        if self.prior is not None:
            prior = torch.index_select(self.prior, 0, targets).unsqueeze(1)
        else:
            prior = 1 / nc
        loss = (1 - prior) / prior * loss_pn / loss_pv
        g_loss = (loss / (1 + loss)).mean()

        return g_loss



import torch
import torch.nn as nn
import torchvision
import re


def densenet121(num_classes=1000, pretrained=False):
        model = torchvision.models.densenet121(pretrained=pretrained)
        num_feats = model.classifier.in_features
        model.classifier = nn.Linear(num_feats, num_classes)
        return model



from .AUROC import AUROC  
from .PAUROC import PartialAUROC
from .AUTKC import TopkAcc, AUTKC
from .OpenAUC import OpenAUC, OpenSetEvaluator, Acc_At_TPR, Acc_At_T, ClosedSetAcc, MicroF, MacroF


# __all__ = ['AUROC', 'PartialAUROC',
#            'TopkAcc', 'AUTKC',
#            'MacroF', 'MicroF',
#            'ClosedSetAcc', 'Acc_At_T',
#            'Acc_At_TPR', 'AUROC', 'OpenAUC', 'OpenSetEvaluator']

__all__ = ['AUROC', 'PartialAUROC',
           'TopkAcc', 'AUTKC',
           'MacroF', 'MicroF',
           'ClosedSetAcc', 'Acc_At_T',
           'Acc_At_TPR', 'OpenAUC', 'OpenSetEvaluator']


import pytest
import sys
sys.path.append("./")

from torch.utils.data import DataLoader
from easydict import EasyDict as edict

from XCurve.AUPRC import RetrievalDataset, RetrievalModel


@pytest.mark.skip()
def load_cfg(dataset):
    if dataset == 'inat':
        from XCurve.AUPRC import DefaultInatDatasetCfg as DefaultDatasetCfg
    elif dataset == 'sop':
        from XCurve.AUPRC import DefaultSOPDatasetCfg as DefaultDatasetCfg
    elif dataset == 'vehid':
        from XCurve.AUPRC import DefaultVehIDDatasetCfg as DefaultDatasetCfg
    else:
        assert NotImplementedError

    return DefaultDatasetCfg

def test_model():
    for dataset_name in ['inat', 'sop', 'vehid']:
        args = load_cfg(dataset_name)
        dataset = RetrievalDataset(**args, split='val')
        dataloader = DataLoader(
            dataset,
            batch_size=args.batchsize,
            num_workers=4
        )

        for frz in [True, False]:
            for frz_bn in [True, False]:
                model = RetrievalModel(512, frz, frz_bn).cuda()
                for i, (img, _) in enumerate(dataloader):
                    img = img.cuda()
                    out = model(img)
                    assert out.shape == (img.shape[0], model.output_channels)                
                    if (i+1)%4 == 0:
                        break


if __name__ == '__main__':
    test_model()



import torch
import numpy as np
import os
from tqdm import tqdm
from sklearn.metrics import roc_curve, roc_auc_score, accuracy_score

# refer to Fig.2 of the paper "Nearest neighbors distance ratio open-set classifier" 
def MacroF(close_pred, close_labels, open_pred, open_labels, t_list):
    ret = list()
    for t in t_list:
        tp, fp, fn = 0, 0, 0
        close_pred_ = np.argmax(close_pred, axis=1)
        open_pred_ = (open_pred > t) * 1
        for c_pred, o_pred, cl, ol in zip(close_pred_, open_pred_, close_labels, open_labels):
            tp += 1 if c_pred == cl and o_pred == 0 and ol == 0 else 0
            fp += 1 if c_pred != cl and o_pred == 0 and ol == 0 else 0
            fp += 1 if o_pred == 0 and ol == 1 else 0
            fn += 1 if c_pred != cl and o_pred == 0 and ol == 0 else 0
            fn += 1 if o_pred == 1 and ol == 0 else 0
        p = tp / (tp + fp) if (tp + fp) > 0 else 0
        r = tp / (tp + fn) if (tp + fn) > 0 else 0
        ret.append(2 * p * r / (p + r) if (p + r) > 0 else 0)
    # m_ret = max(ret)
    # print(f'M-F-score: {m_ret}', ret)
    return ret

def MicroF(close_pred, close_labels, open_pred, open_labels, t_list):
    ret = list()
    n_class = close_pred.shape[1]
    for t in t_list:
        tp, fp, fn = [0, ] * n_class, [0, ] * n_class, [0, ] * n_class
        close_pred_ = np.argmax(close_pred, axis=1)
        open_pred_ = (open_pred > t) * 1
        for c_pred, o_pred, cl, ol in zip(close_pred_, open_pred_, close_labels, open_labels):
            tp[cl] += 1 if c_pred == cl and o_pred == 0 and ol == 0 else 0
            fp[c_pred] += 1 if c_pred != cl and o_pred == 0 and ol == 0 else 0
            fp[c_pred] += 1 if o_pred == 0 and ol == 1 else 0
            fn[cl] += 1 if c_pred != cl and o_pred == 0 and ol == 0 else 0
            fn[cl] += 1 if o_pred == 1 and ol == 0 else 0
        p = sum([tpi / (tpi + fpi) if (tpi + fpi) > 0 else 0 for tpi, fpi in zip(tp, fp)]) / n_class
        r = sum([tpi / (tpi + fni) if (tpi + fni) > 0 else 0 for tpi, fni in zip(tp, fn)]) / n_class
        ret.append(2 * p * r / (p + r) if (p + r) > 0 else 0)
    # m_ret = max(ret)
    # print(f'm-F-score: {m_ret}', ret)
    return ret

def ClosedSetAcc(preds, labels):
    preds = preds.argmax(axis=-1)
    acc = accuracy_score(labels, preds)
    # print('Closed Set Accuracy: {:.3f}'.format(acc))
    return acc

def find_nearest(array, value):
    array = np.asarray(array)
    length = len(array)
    abs_diff = np.abs(array - value)

    t_star = abs_diff.min()
    equal_arr = (abs_diff == t_star).astype('float32') + np.linspace(start=0, stop=0.1, num=length)

    idx = equal_arr.argmax()

    return array[idx], idx

def Acc_At_T(preds, labels, t):
    pred_t = np.copy(preds)
    pred_t[pred_t > t] = 1
    pred_t[pred_t <= t] = 0

    acc = accuracy_score(labels, pred_t.astype('int32'))

    return acc

def Acc_At_TPR(open_set_preds, open_set_labels, r=0.95):
    _, tpr, thresholds = roc_curve(open_set_labels, open_set_preds, drop_intermediate=False)

    # Error rate at r TAR
    _, idx = find_nearest(tpr, r)
    t = thresholds[idx]
    acc = Acc_At_T(open_set_preds, open_set_labels, t)
    return acc

def AUROC(open_set_preds, open_set_labels):
    auroc = roc_auc_score(open_set_labels, open_set_preds)
    # print(f'AUROC: {auroc}')
    return auroc

def OpenAUC(open_set_pred_known, open_set_pred_unknown, close_set_pred_class, close_set_labels):
    """
    :param open_set_pred_known: open set score for each known class sample (B_k,)
    :param open_set_pred_unknown: open set score for each unknown class sample (B_u,)
    :param close_set_pred_class: predicted class for each known class sample (B_k,)
    :param close_set_labels: correct class for each known class sample (B_k,)
    :return: OpenAUC
    """
    open_set_pred_known, open_set_pred_unknown, correct = open_set_pred_known.tolist(), open_set_pred_unknown.tolist(), (close_set_pred_class == close_set_labels).tolist()
    m_x2 = max(open_set_pred_unknown) + 1e-5
    y_score = [value if hit else m_x2 for value, hit in zip(open_set_pred_known, correct)] + open_set_pred_unknown
    y_true = [0] * len(open_set_pred_known) + [1] * len(open_set_pred_unknown)
    open_auc = roc_auc_score(y_true, y_score)
    # print('OpenAUC:', open_auc)
    return open_auc

class EnsembleModel:
    def __init__(self, model, use_softmax=False):
        self.model = model
        self.use_softmax = use_softmax

    def forward(self, imgs):
        """
        :param imgs:
        :return: Closed set and open set predictions on imgs
        """
        _, closed_set_preds = self.model(imgs, True)
        if self.use_softmax:
            closed_set_preds = torch.nn.Softmax(dim=-1)(closed_set_preds)

        open_set_preds = - closed_set_preds.max(dim=-1)[0]
        open_set_preds = torch.nn.Sigmoid()(open_set_preds)

        return closed_set_preds, open_set_preds

class OpenSetEvaluator():

    def __init__(self, model: EnsembleModel, known_data_loader, unknown_data_loader, save_dir=None, device=None):

        self.model = model
        self.known_data_loader = known_data_loader
        self.unknown_data_loader = unknown_data_loader
        self.save_dir = save_dir

        if device is None:
            device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        self.device = device

        # Init empty lists for saving labels and preds
        self.closed_set_preds = {0: [], 1: []}
        self.open_set_preds = {0: [], 1: []}

        self.closed_set_labels = {0: [], 1: []}
        self.open_set_labels = {0: [], 1: []}

        if save_dir and not os.path.exists(save_dir):
            os.mkdir(save_dir)

    def predict(self, save=True):

        with torch.no_grad():
            for open_set_label, loader in enumerate((self.known_data_loader, self.unknown_data_loader)):

                if open_set_label:
                    print('Forward pass through Open Set test set...')
                else:
                    print('Forward pass through Closed Set test set...')

                for batch_idx, batch in enumerate(tqdm(loader)):

                    imgs, labels, idxs = [x.to(self.device) for x in batch]

                    # Model forward
                    output = self.model(imgs)
                    closed_set_preds, open_set_preds = [x.cpu().numpy().tolist() for x in output]

                    # Update preds and labels
                    self.closed_set_preds[open_set_label].extend(closed_set_preds)
                    self.open_set_preds[open_set_label].extend(open_set_preds)

                    self.closed_set_labels[open_set_label].extend(labels.cpu().numpy().tolist())
                    self.open_set_labels[open_set_label].extend([open_set_label] * len(labels))

        if save: # Save to disk
            save_names = ['closed_set_preds.pt', 'open_set_preds.pt', 'closed_set_labels.pt', 'open_set_labels.pt']
            save_lists = [self.closed_set_preds, self.open_set_preds, self.closed_set_labels, self.open_set_labels]

            for name, x in zip(save_names, save_lists):

                path = os.path.join(self.save_dir, name)
                torch.save(x, path)
        else:
            return self.closed_set_preds, self.open_set_preds, self.closed_set_labels, self.open_set_labels

    @staticmethod
    def evaluate(self, load=True, preds=None):

        if load:
            save_names = ['closed_set_preds.pt', 'open_set_preds.pt', 'closed_set_labels.pt', 'open_set_labels.pt']

            closed_set_preds, open_set_preds, closed_set_labels, open_set_labels = \
                [torch.load(os.path.join(self.save_dir, name)) for name in save_names]

        else:

            closed_set_preds, open_set_preds, closed_set_labels, open_set_labels = preds

        open_set_preds = np.array(open_set_preds[0] + open_set_preds[1])
        open_set_labels = np.array(open_set_labels[0] + open_set_labels[1])

        # ----------------------------
        # CLOSED SET EVALUATION
        # ----------------------------

        close_set_acc = ClosedSetAcc(np.array(closed_set_preds[0]), np.array(closed_set_labels[0]))

        # ----------------------------
        # OPEN SET EVALUATION
        # ----------------------------

        acc_95 = Acc_At_TPR(open_set_preds, open_set_labels)
        auroc = AUROC(open_set_preds, open_set_labels)

        open_set_preds_known_cls = open_set_preds[~open_set_labels.astype('bool')]
        open_set_preds_unknown_cls = open_set_preds[open_set_labels.astype('bool')]
        closed_set_preds_pred_cls = np.array(closed_set_preds[0]).argmax(axis=-1)
        labels_known_cls = np.array(closed_set_labels[0])

        open_auc = OpenAUC(open_set_preds_known_cls, open_set_preds_unknown_cls, closed_set_preds_pred_cls, labels_known_cls)

        t_list = [_ * 0.01 for _ in range(1, 100)]
        closed_set_preds = np.array(closed_set_preds[0] + closed_set_preds[1])
        closed_set_labels = np.array(closed_set_labels[0] + closed_set_labels[1])
        macroF = MacroF(closed_set_preds, closed_set_labels, open_set_preds, open_set_labels, t_list)
        microF = MicroF(closed_set_preds, closed_set_labels, open_set_preds, open_set_labels, t_list)

        results = {
            'Acc': close_set_acc,
            'Acc_95': acc_95,
            'AUROC': auroc,
            'OpenAUC': open_auc,
            'macroF': macroF,
            'microF': microF,
        }

        return results


import pytest
import random
import sys
from torch.utils.data import DataLoader
sys.path.append("./")

from XCurve.AUPRC import RetrievalDataset


@pytest.mark.skip()
def load_cfg(dataset):
    if dataset == 'inat':
        from XCurve.AUPRC import DefaultInatDatasetCfg as DefaultDatasetCfg
    elif dataset == 'sop':
        from XCurve.AUPRC import DefaultSOPDatasetCfg as DefaultDatasetCfg
    elif dataset == 'vehid':
        from XCurve.AUPRC import DefaultVehIDDatasetCfg as DefaultDatasetCfg
    else:
        assert NotImplementedError

    return DefaultDatasetCfg

def test_single_dataset():
    for dataset_name in ['inat', 'sop', 'vehid']:
        args = load_cfg(dataset_name)
        dataset = RetrievalDataset(**args)

        for rep in range(10):
            idx = random.randint(0, len(dataset) - 1)
            img, lbl = dataset.__getitem__(idx)
        assert img.shape == (3, args.input_size, args.input_size)
        assert lbl.shape == (1,)

def test_dataloader():

    for dataset_name in ['inat', 'sop', 'vehid']:
        args = load_cfg(dataset_name)

        for subset in ['train', 'val', 'test']:
            dataset = RetrievalDataset(**args, split=subset)
            dataloader = DataLoader(
                dataset,
                batch_size=args.batchsize,
                num_workers=4
            )

            for i, (img, lbl) in enumerate(dataloader):
                assert img.shape[0] == lbl.shape[0] == 56
                assert img.shape[1:] == (3, args.input_size, args.input_size)
                assert len(lbl.shape) == 2
                assert lbl.shape[1] == 1
                if (i+1)%10 == 0:
                    break

            assert len(dataloader.dataset.get_cnt_per_id()) > 0
            dataloader.dataset.reset()


if __name__ == '__main__':
    test_single_dataset()
    test_dataloader()



import torch
import torchvision
import torchvision.transforms as transforms
import numpy as np
import torchvision.datasets as datasets
from sampler import *
import warnings
import os
from PIL import Image
import numpy as np


class IMBALANCEMNIST(torchvision.datasets.MNIST):
    cls_num = 10
    def __init__(self, root, imb_type='exp', imb_factor=0.01, rand_number=0, train=True,
                 transform=None, target_transform=None,
                 download=False):
        super(IMBALANCEMNIST, self).__init__(root, train, transform, target_transform, download)
        np.random.seed(rand_number)
        img_num_list = self.get_img_num_per_cls(self.cls_num, imb_type, imb_factor)
        self.phat = 0.1
        self.gen_imbalanced_data(img_num_list)

    def __getitem__(self, index):
        """
		Args:
			index (int): Index

		Returns:
			tuple: (image, target) where target is index of the target class.
		"""
        img, target = self.data[index], int(self.targets[index])
    
        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img, mode='L')
    
        if self.transform is not None:
            img = self.transform(img)
    
        if self.target_transform is not None:
            target = self.target_transform(target)
    
        return img, target
        
    def get_img_num_per_cls(self, cls_num, imb_type, imb_factor):
        targets_np = np.array(self.targets, dtype=np.int64)
        # img_max = len(self.data) / cls_num
        img_num_per_cls = []
        
        for cls_idx in range(cls_num):
            img_max = np.where(targets_np == cls_idx)[0].__len__()
            num = img_max * (imb_factor ** (cls_idx / (cls_num - 1.0)))
            img_num_per_cls.append(int(num))
        
        return img_num_per_cls
    
    def gen_imbalanced_data(self, img_num_per_cls):
        new_data = []
        new_targets = []
        targets_np = np.array(self.targets, dtype=np.int64)
        classes = np.unique(targets_np)
        # np.random.shuffle(classes)
        self.num_per_cls_dict = dict()
        for the_class, the_img_num in zip(classes, img_num_per_cls):
            # print('need', the_class, the_img_num)
            self.num_per_cls_dict[the_class] = the_img_num
            idx = np.where(targets_np == the_class)[0]
            np.random.shuffle(idx)
            selec_idx = idx[:the_img_num]
            # print(len(selec_idx))
            new_data.append(self.data[selec_idx, ...])
            new_targets.extend([the_class, ] * the_img_num)
        new_data = np.vstack(new_data)
        new_targets = self.get_two_class(new_targets)
        self.data = new_data
        self.targets = new_targets
        # print(len(self.targets))
        # print(self.data.shape)
    
    def get_cls_num_list(self):
        cls_num_list = []
        for i in range(self.cls_num):
            cls_num_list.append(self.num_per_cls_dict[i])
        return cls_num_list
    
    def get_two_class(self, Y):
        Y = np.array(Y)
        loc_0 = np.where(Y <= 4)[0]
        loc_1 = np.where(Y > 4)[0]
        Y[loc_1] = 1
        Y[loc_0] = 0
        self.phat = len(np.where(Y == 1)[0]) / len(Y)
        return Y.tolist()
    
if __name__=='__main__':
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
    ])
    
    train_dataset = IMBALANCEMNIST(root='../', train=True, download=True, transform=transform_train)
    
    train_sampler = StratifiedSampler(class_vector=train_dataset.targets,
                                      batch_size=128,
                                      rpos=1,
                                      rneg=9)

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=128, shuffle=(train_sampler is None),
        num_workers=2,
        # pin_memory=True,
        sampler=train_sampler)
    
    for i, (x, y) in enumerate(train_loader):
        print(i, x.min(), x.max(), x.mean(), x.std())


import torch
import torch.nn as nn
from abc import ABC


class BaseModel(nn.Module, ABC):
    def __init__(self):
        super().__init__()

    def freeze(self):
        for p in self.parameters():
            p.requires_grad = False

    def freezeBN(self):
        for m in self.modules():
            class_name = m.__class__.__name__
            if class_name.find('BatchNorm') != -1:
                m.eval()
                m.train = lambda _:None

    def param_groups(self, lr=None, **kwargs):
        params = list(filter(lambda x:x.requires_grad, self.parameters()))

        if len(params):
            if lr is not None:
                return [{'params': params, 'lr': lr}]
            else:
                return [{'params': params}]
        else:
            return []



import yaml
import os
import sys
sys.path.append("./")

from easydict import EasyDict as edict

PWD = os.path.dirname(os.path.abspath(__file__))

def load_yaml(filepath):
    with open(filepath, 'r') as f:
        args = yaml.load(f, Loader=yaml.FullLoader)
    return edict(args)

DefaultLossCfg = load_yaml(PWD + '/loss.yaml')
DefaultInatDatasetCfg = load_yaml(PWD + '/datasets/inaturalist.yaml')
DefaultSOPDatasetCfg = load_yaml(PWD + '/datasets/sop.yaml')
DefaultVehIDDatasetCfg = load_yaml(PWD + '/datasets/vehid.yaml')



# coding=utf-8
import torch
import torch.nn as nn

class AdvAUROCLoss(nn.Module):
    def __init__(self, imratio=None,
                 a=None,
                 b=None,
                 alpha=None):
                         
        super(AdvAUROCLoss, self).__init__()
        
        self.p = imratio
        if a is not None:
            self.a = torch.tensor(a).float().cuda()
            self.a.requires_grad = True
        else:
            self.a = torch.tensor(0.2).float().cuda()
            self.a.requires_grad = True
            # self.a = torch.zeros(1, dtype=torch.float32, requires_grad=True).cuda()
        
        if b is not None:
            self.b = torch.tensor(b).float().cuda()
            self.b.requires_grad = True
        else:
            # self.b = torch.zeros(1, dtype=torch.float32, requires_grad=True).cuda()
            self.b = torch.tensor(0.2).float().cuda()
            self.b.requires_grad = True
        
        if alpha is not None:
            self.alpha = torch.tensor(alpha).float().cuda()
            self.alpha.requires_grad = True
        else:
            # self.alpha = torch.zeros(1, dtype=torch.float32, requires_grad=True).cuda()
            self.alpha = torch.tensor(0.2).float().cuda()
            self.alpha.requires_grad = True
    
    def forward(self, y_pred, y_true):
        if self.p is None:
            self.p = (y_true == 1).float().sum() / y_true.shape[0]
        
        y_pred = y_pred.reshape(-1, 1)
        y_true = y_true.reshape(-1, 1)
        loss = (1 - self.p) * torch.mean((y_pred - self.a) ** 2 * (1 == y_true).float()) + \
               self.p * torch.mean((y_pred - self.b) ** 2 * (0 == y_true).float()) + \
               2 * self.alpha * (self.p * (1 - self.p) +
                                 torch.mean((self.p * y_pred * (0 == y_true).float() - (1 - self.p) * y_pred * (
                                             1 == y_true).float()))) - \
               self.p * (1 - self.p) * self.alpha ** 2
        return loss
    

class RegAdvAUROCLoss(AdvAUROCLoss):
    def __init__(self, imratio=None,
                 a=None,
                 b=None,
                 alpha=None,
                 lambda1=None,
                 lambda2=None) -> None:
        super(RegAdvAUROCLoss, self).__init__()
        if lambda1 is not None:
            self.lambda1 = torch.tensor(lambda1).float().cuda()
            self.lambda1.requires_grad = True
        else:
            self.lambda1 = torch.tensor(1.0).float().cuda()
            self.lambda1.requires_grad = True
        
        if lambda2 is not None:
            self.lambda2 = torch.tensor(lambda2).float().cuda()
            self.lambda2.requires_grad = True
        else:
            self.lambda2 = torch.tensor(1.0).float().cuda()
            self.lambda2.requires_grad = True
    
    def forward(self, y_pred, y_true):
        if self.p is None:
            self.p = (y_true == 1).float().sum() / y_true.shape[0]
        
        y_pred = y_pred.reshape(-1, 1)
        y_true = y_true.reshape(-1, 1)
        
        loss = self.get_loss(y_pred, y_true)
        loss = loss - self.lambda1*(self.alpha + self.a) - self.lambda2*(self.alpha - self.b + 1)
        return loss

class selfloss(nn.Module):
    def __init__(self):
        super(selfloss, self).__init__()
    
    def forward(self, y_pred, y_true):
        y_pred = y_pred.reshape(-1, 1)
        y_true = y_true.reshape(-1, 1)
        # new_y = y_true
        # new_y[new_y==0] = -1
        loss = torch.mean(-1 * (1 == y_true).float() * y_pred + (0 == y_true).float() * y_pred)
        return loss